/* ====================================================================
*
* The ObjectStyle Group Software License, version 1.1
* ObjectStyle Group - http://objectstyle.org/
* 
* Copyright (c) 2002-2005, Andrei (Andrus) Adamchik and individual authors
* of the software. All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in
*    the documentation and/or other materials provided with the
*    distribution.
* 
* 3. The end-user documentation included with the redistribution, if any,
*    must include the following acknowlegement:
*    "This product includes software developed by independent contributors
*    and hosted on ObjectStyle Group web site (http://objectstyle.org/)."
*    Alternately, this acknowlegement may appear in the software itself,
*    if and wherever such third-party acknowlegements normally appear.
* 
* 4. The names "ObjectStyle Group" and "Cayenne" must not be used to endorse
*    or promote products derived from this software without prior written
*    permission. For written permission, email
*    "andrus at objectstyle dot org".
* 
* 5. Products derived from this software may not be called "ObjectStyle"
*    or "Cayenne", nor may "ObjectStyle" or "Cayenne" appear in their
*    names without prior written permission.
* 
* THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED.  IN NO EVENT SHALL THE OBJECTSTYLE GROUP OR
* ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
* ====================================================================
* 
* This software consists of voluntary contributions made by many
* individuals and hosted on ObjectStyle Group web site.  For more
* information on the ObjectStyle Group, please see
* <http://objectstyle.org/>.
*/
package org.objectstyle.cayenne.dataview;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.net.URL;
import java.text.Format;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.math.NumberUtils;
import org.apache.crimson.tree.TextNode;
import org.objectstyle.cayenne.CayenneRuntimeException;
import org.objectstyle.cayenne.DataObject;
import org.objectstyle.cayenne.access.DataContext;
import org.objectstyle.cayenne.conf.Configuration;
import org.objectstyle.cayenne.map.EntityResolver;
import org.objectstyle.cayenne.map.ObjAttribute;
import org.objectstyle.cayenne.map.ObjEntity;
import org.objectstyle.cayenne.map.ObjRelationship;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Provides the root of the Data View configuration hierarchy. A <tt>DataView</tt> 
 * contains a number of <tt>ObjEntityView</tt> objects.
 * <p/>
 * An example of creating a DataView is provided below:
 * 
 * <pre style="border:1px dashed gray;padding:1em;">
 * DataView dataView = new DataView();
 *		
 * DataContext dataContext = DataContext.createDataContext();
 *
 * dataView.setEntityResolver(dataContext.getEntityResolver());
 *
 * Configuration.getSharedConfiguration().loadDataView(dataView);  </pre>
 * 
 * Alernatively you can invoke the {@link #createDataView()} factory 
 * method.
 * 
 * <pre style="border:1px dashed gray;padding:1em;">
 * DataView dataView = DataView.createDataView();  </pre>
 * 
 * @see ObjEntityView
 * 
 * @since 1.1
 * @author Andriy Shapochka
 * @author Malcolm Edgar
 */
public class DataView {

    /** Data type definitions. */
	protected DataTypeSpec dataTypeSpec = new DataTypeSpec();
    
    /** ObjEntity lookup EntityResolver. */
	protected EntityResolver entityResolver;
    
    /** The field value change EventDispatcher. */
	protected EventDispatcher fieldValueChangeDispatcher;
    
    /** Format definitions. */
	protected FormatFactory formatFactory = new FormatFactory();
    
    /** The set ObjEntityViewField lookup fields. */
    protected Set lookupEntityViewFields = new HashSet();
    
    /** 
     * The application locale, used to create the format objects. The default
     * value is the default <tt>Locale</tt>. 
     */
    protected Locale locale = Locale.getDefault();
    
    /** The lookup cache for <tt>ObjEntityViewField</tt> objects. */
    protected LookupCache lookupCache = new LookupCache();

    /** The ordered map of ObjEntityView objects. */
    protected Map objEntityViews = new TreeMap();
    
    /** A holder of a DataView bound to the current thread. */
    protected static final ThreadLocal threadDataView = new ThreadLocal();

    // --------------------------------------------------------- Public Methods
    
    /**
     * Factory method that creates and returns a new instance of DataView based 
     * on default domain and shared configuration.
     * 
     * @return a new DataView instance based on the shared configuration
     * @throws CayenneRuntimeException if a configuration error occurs
     */
    public static DataView createDataView() {
    	DataView dataView = new DataView();
		
		DataContext dataContext = DataContext.createDataContext();

		dataView.setEntityResolver(dataContext.getEntityResolver());
		
		try {
			Configuration.getSharedConfiguration().loadDataView(dataView);

			return dataView;
			
		} catch (IOException ioe) {
			throw new CayenneRuntimeException(ioe);
		}
    }

    /**
     * Returns the DataView bound to the current thread.
     * 
     * @return the DataView associated with caller thread.
     * @throws IllegalStateException if there is no DataView bound to the current
     *             thread.
     */
    public static DataView getThreadDataView() throws IllegalStateException {
    	DataView dv = (DataView) threadDataView.get();
        if (dv == null) {
            throw new IllegalStateException("Current thread has no bound DataView.");
        }

        return dv;
    }
    
    /**
     * Return true if a DataView is bound to the current thread.
     * 
     * @return true if a DataView is bound to the current thread
     */
    public static boolean hasThreadDataView() {
    	return (threadDataView.get() != null);
    }

    /**
     * Binds a DataView to the current thread. DataView can later be retrieved by
     * users in the same thread by calling {@link DataView#getThreadDataView()}.
     * Using null parameter will unbind currently bound DataView.
     * 
     * @param dataView the DataView to bind the the current thread
     */
    public static void bindThreadDataView(DataView dataView) {
    	threadDataView.set(dataView);
    }

    /**
     * Return the Cayenne EntityResolver for this object.
     * 
     *  @return the Cayenne EntityResolver for this object
     */
    public EntityResolver getEntityResolver() {
        return entityResolver;
    }

    /**
     * Set the Cayenne EntityResolver for this object.
     * 
     * @param entityResolver the Cayenne EntityResolver for this object
     */
    public void setEntityResolver(EntityResolver entityResolver) {
        Validate.notNull(entityResolver);
        this.entityResolver = entityResolver;
    }

    /**
     * Load the DataView from the given list of DataView XML source files.
     * 
     * @param xmlSources the DataView XML source files to load
     * @throws CayenneRuntimeException if an error occurs
     */
    public void load(File[] xmlSources) {
        Validate.noNullElements(xmlSources);

        Document[] documents = new Document[xmlSources.length];
        
        DocumentBuilderFactory factory =
            DocumentBuilderFactory.newInstance();
        
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
        
            for (int i = 0; i < xmlSources.length; i++) {
            	documents[i] = builder.parse(xmlSources[i]);
            }
            
            load(documents);
            
        } catch (IOException ioe) {
        	throw new CayenneRuntimeException(ioe);
           
        } catch (SAXException se) {
        	throw new CayenneRuntimeException(se);
        	
        } catch (ParserConfigurationException pce) {
            throw new CayenneRuntimeException(pce);
        }
    }

    /**
     * Load the DataView from the given list of DataView XML input sources.
     * 
     * @param xmlSources the DataView XML input sources to load
     * @throws CayenneRuntimeException if an error occurs
     */
    public void load(InputSource[] xmlSources) throws IOException {
        Validate.noNullElements(xmlSources);

        Document[] documents = new Document[xmlSources.length];
        
        DocumentBuilderFactory factory =
            DocumentBuilderFactory.newInstance();
        
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
        
            for (int i = 0; i < xmlSources.length; i++) {
            	documents[i] = builder.parse(xmlSources[i]);
            }
            
            load(documents);
            
        } catch (IOException ioe) {
        	throw new CayenneRuntimeException(ioe);
           
        } catch (SAXException se) {
        	throw new CayenneRuntimeException(se);
        	
        } catch (ParserConfigurationException pce) {
            throw new CayenneRuntimeException(pce);
        }
    }

    /**
     * Load the DataView from the given list of DataView XML input streams.
     * 
     * @param xmlSources the DataView XML input streams to load
     * @throws CayenneRuntimeException if an error occurs
     */
    public void load(InputStream[] xmlSources) throws IOException {
        Validate.noNullElements(xmlSources);
        
        InputSource[] inputSources = new InputSource[xmlSources.length];
        
        for (int i = 0; i < xmlSources.length; i++) {
        	inputSources[i] = new InputSource(xmlSources[i]);
        }
        
        load(inputSources);
    }

    /**
     * Load the DataView from the given list of DataView XML reader sources.
     * 
     * @param xmlSources the DataView XML reader sources to load
     * @throws CayenneRuntimeException if an error occurs
     */
    public void load(Reader[] xmlSources) throws IOException {
        Validate.noNullElements(xmlSources);
        
        InputSource[] inputSources = new InputSource[xmlSources.length];
        
        for (int i = 0; i < xmlSources.length; i++) {
        	inputSources[i] = new InputSource(xmlSources[i]);
        }
        
        load(inputSources);
    }

    /**
     * Load the DataView from the given list of DataView XML URL sources.
     * 
     * @param xmlSources the DataView XML URL sources to load
     * @throws CayenneRuntimeException if an error occurs
     */
    public void load(URL[] xmlSources) throws IOException {
        Validate.noNullElements(xmlSources);
        
        InputSource[] inputSources = new InputSource[xmlSources.length];
        
        for (int i = 0; i < xmlSources.length; i++) {
        	String systemID = xmlSources[i].toExternalForm();
        	inputSources[i] = new InputSource(systemID);
        }
        
        load(inputSources);
    }

    /**
     * Load the DataView XML documents. This method will resolve any field
     * lookup references.
     * 
     * @param views the DataView XML documents
     */
    public void load(Document[] views) {
        Validate.noNullElements(views);
        
        Map lookupReferenceTable = new HashMap();
        
        for (int i = 0; i < views.length; i++) {
            Element root = views[i].getDocumentElement();
            List entityViews = getChildren(root, "obj-entity-view");
            for (Iterator j = entityViews.iterator(); j.hasNext();) {
                Element entityViewElement = (Element) j.next();
                loadEntityView(entityViewElement, lookupReferenceTable);
            }
        }
        
        resolveLookupReferences(lookupReferenceTable);
    }

    /**
     * Return an unmodifiable set of <tt>ObjEntityView</tt> names.
     * 
     * @return an unmodifiable set of <tt>ObjEntityView</tt> names
     */
    public Set getObjEntityViewNames() {
        return Collections.unmodifiableSet(objEntityViews.keySet());
    }

    /**
     * Return an unmodifiable collection of <tt>ObjEntityView</tt> objects
     * 
     * @return an unmodifiable collection of <tt>ObjEntityView</tt> objects
     */
    public Collection getObjEntityViews() {
        return Collections.unmodifiableCollection(objEntityViews.values());
    }

    /**
     * Return the <tt>ObjEntityView</tt> for the given name.
     * 
     * @param viewName the name of the <tt>ObjEntityView</tt>
     * @return the <tt>ObjEntityView</tt> for the given name
     */
    public ObjEntityView getObjEntityView(String viewName) {
        return (ObjEntityView) objEntityViews.get(viewName);
    }

    /**
     * Return the <tt>ObjEntityViewField</tt> lookup cache.
     * 
     * @return the <tt>ObjEntityViewField</tt> lookup cache
     */
    public LookupCache getLookupCache() {
        return lookupCache;
    }

    /**
     * Set the <tt>ObjEntityViewField</tt> lookup cache.
     * 
     * @param lookupCache the <tt>ObjEntityViewField</tt> lookup cache
     */
    public void setLookupCache(LookupCache lookupCache) {
        Validate.notNull(lookupCache);
        this.lookupCache = lookupCache;
    }

    public Set getLookupObjEntityViewFields() {
        return Collections.unmodifiableSet(lookupEntityViewFields);
    }

    /**
     * Return the application locale.
     * 
     * @return the applicaton locale
     */
    public Locale getLocale() {
        return locale;
    }

    /**
     * Set the application locale.
     * 
     * @param locale the applicaton locale
     */
    public void setLocale(Locale locale) {
        this.locale = locale;
    }

    /**
     * Return the <tt>DataTypeSpec</tt>.
     * 
     * @return the <tt>DataTypeSpec</tt>
     */
    public DataTypeSpec getDataTypeSpec() {
        return dataTypeSpec;
    }


    /**
     * Set the <tt>DataTypeSpec</tt>.
     * 
     * @param dataTypeSpec the <tt>DataTypeSpec</tt>
     */
    public void setDataTypeSpec(DataTypeSpec dataTypeSpec) {
        this.dataTypeSpec = dataTypeSpec;
    }

    /**
     * Return the <tt>FormatFactory</tt>.
     * 
     * @return the <tt>FormatFactory</tt>
     */
    public FormatFactory getFormatFactory() {
        return formatFactory;
    }

    /**
     * Set the <tt>FormatFactory</tt>.
     * 
     * @param formatFactory the <tt>FormatFactory</tt> to set
     */
    public void setFormatFactory(FormatFactory formatFactory) {
        this.formatFactory = formatFactory;
    }

    /**
     * Add a <tt>DataObject</tt> field value change listener.
     * 
     * @param listener a <tt>DataObject</tt> field value change listener to add
     */
    public void addFieldValueChangeListener(FieldValueChangeListener listener) {
        fieldValueChangeDispatcher = EventDispatcher.add(
                fieldValueChangeDispatcher,
                listener);
    }

    /**
     * Remove a <tt>DataObject</tt> field value change listener.
     * 
     * @param listener a <tt>DataObject</tt> field value change listener to remove
     */
    public void removeFieldValueChangeListener(FieldValueChangeListener listener) {
        fieldValueChangeDispatcher = EventDispatcher.remove(
                fieldValueChangeDispatcher,
                listener);
    }

    /**
     * Clear the <tt>DataObject</tt> field value change listeners.
     */
    public void clearFieldValueChangeListeners() {
        if (fieldValueChangeDispatcher != null) {
            fieldValueChangeDispatcher.clear();
            fieldValueChangeDispatcher = null;
        }
    }

    /**
     * Fire a <tt>FieldValueChangeEvent</tt> for the given <tt>DataObject</tt>
     * and <tt>ObjEntityViewField</tt>
     * 
     * @param source the source <tt>ObjEntityViewField</tt>
     * @param modifiedObject the modified <tt>DataObject</tt>
     * @param oldValue the old field value
     * @param newValue the new field value
     */
    public void fireFieldValueChangeEvent(
            ObjEntityViewField source,
            DataObject modifiedObject,
            Object oldValue,
            Object newValue) {
    	
        if (fieldValueChangeDispatcher != null && source.getRootOwner() == this) {
        	
            fieldValueChangeDispatcher.dispatch(new FieldValueChangeEvent(
                    source,
                    modifiedObject,
                    oldValue,
                    newValue));
        }
    }

    // -------------------------------------------------------- Private Methods
    
    /**
     * Resolve the loaded lookup field references for the ObjEntityViewField
     * objects.
     * 
     * @param lookupReferenceTable the Map of field lookup references
     */
    private void resolveLookupReferences(Map lookupReferenceTable) {
    	
        lookupEntityViewFields.clear();
        
        for (Iterator i = lookupReferenceTable.entrySet().iterator(); i.hasNext();) {
        	
            Map.Entry entry = (Map.Entry) i.next();
            ObjEntityViewField field = (ObjEntityViewField) entry.getKey();
            String[] lookupDescriptor = (String[]) entry.getValue();
            
            ObjEntityView lookupEntityView = getObjEntityView(lookupDescriptor[0]);
            ObjEntityViewField lookupField = 
            	lookupEntityView.getField(lookupDescriptor[1]);
            
            field.setLookupField(lookupField);
            
            lookupEntityViewFields.add(lookupField);
        }
    }

    /**
     * Load the given ObjEntityView XML element into this DataView.
     * 
     * @param element the ObjEntityView's XML element
     * @param lookupReferenceTable the Map of field lookup references
     */
    private void loadEntityView(Element element, Map lookupReferenceTable) {
    	
        String name = element.getAttribute("name");
        
        Validate.isTrue(name != null && !objEntityViews.containsKey(name));
        
        String objEntityName = element.getAttribute("obj-entity-name");
        Validate.notNull(objEntityName);
        
        ObjEntity objEntity = entityResolver.getObjEntity(objEntityName);
        
        ObjEntityView entityView = new ObjEntityView();
        entityView.setName(name);
        entityView.setObjEntity(objEntity);
        objEntityViews.put(name, entityView);
        entityView.setOwner(this);
        
        List fields = getChildren(element, "field");
        for (Iterator i = fields.iterator(); i.hasNext();) {
            Element fieldElement = (Element) i.next();
            loadField(entityView, fieldElement, lookupReferenceTable);
        }
    }

    /**
     * Load the given fields XML element into the given ObjEntityView.
     * 
     * @param entityView the entity view to load the field into
     * @param element the field's XML element description
     * @param lookupReferenceTable the Map of field lookup references
     */
    private void loadField(ObjEntityView entityView, Element element, Map lookupReferenceTable) {
    	
        String name = element.getAttribute("name");
        
        ObjEntityViewField field = new ObjEntityViewField();
        field.setName(name);
        String prefIndex = element.getAttribute("pref-index");
        field.setPreferredIndex(NumberUtils.toInt(prefIndex, -1));
        entityView.insertField(field);

        String calcType = element.getAttribute("calc-type");
        Validate.notNull(calcType);
        CalcTypeEnum fieldCalcType = CalcTypeEnum.getEnum(calcType);
        Validate.isTrue(
                CalcTypeEnum.NO_CALC_TYPE.equals(fieldCalcType)
                        || CalcTypeEnum.LOOKUP_TYPE.equals(fieldCalcType),
                "Calc Type not supported yet: ",
                fieldCalcType);
        field.setCalcType(fieldCalcType);

        ObjEntity objEntity = entityView.getObjEntity();

        if (CalcTypeEnum.NO_CALC_TYPE.equals(fieldCalcType)) {
        	
            String objAttributeName = element.getAttribute("obj-attribute-name");
            Validate.notNull(objAttributeName);
            
            ObjAttribute objAttribute = (ObjAttribute) 
            	objEntity.getAttribute(objAttributeName);
            
            field.setObjAttribute(objAttribute);
            
        } else if (CalcTypeEnum.LOOKUP_TYPE.equals(fieldCalcType)) {
        	
            String objRelationshipName = element.getAttribute("obj-relationship-name");
            Validate.notNull(objRelationshipName);
            
            ObjRelationship objRelationship = (ObjRelationship) 
            	objEntity.getRelationship(objRelationshipName);
            
            field.setObjRelationship(objRelationship);
            Element lookupElement = getChild(element, "lookup");
            Validate.notNull(lookupElement);
            
            String lookupEntityView = 
            	lookupElement.getAttribute("obj-entity-view-name");
            Validate.notNull(lookupEntityView);
            
            String lookupEntityField = lookupElement.getAttribute("field-name");
            Validate.notNull(lookupEntityField);
            
            String[] lookupDescriptor = new String[] {
            		lookupEntityView, lookupEntityField
            };
            
            lookupReferenceTable.put(field, lookupDescriptor);
        }

        String dataType = element.getAttribute("data-type");
        Validate.notNull(dataType);
        field.setDataType(getDataTypeSpec().getDataType(dataType));

        String editable = element.getAttribute("editable");
        field.setEditable(BooleanUtils.toBoolean(editable));

        String visible = element.getAttribute("visible");
        field.setVisible(BooleanUtils.toBoolean(visible));

        Element captionElement = getChild(element, "caption");
        if (captionElement != null) {
        	String caption = getElementText(captionElement);
            field.setCaption(caption);
        }

        Element editFormatElement = getChild(element, "edit-format");
        if (editFormatElement != null) {
            String formatClassName = editFormatElement.getAttribute("class");
            Validate.notNull(formatClassName);

            try {
                Class formatClass = Class.forName(formatClassName);
                Map parameters = childrenToMap(editFormatElement);
                
                Format format = getFormatFactory().createFormat(
                        formatClass,
                        getLocale(),
                        parameters);
                
                field.setEditFormat(format);
                
            } catch (ClassNotFoundException cnfe) {
            	throw new CayenneRuntimeException(cnfe);
            }
        }

        Element displayFormatElement = getChild(element, "display-format");
        if (displayFormatElement != null) {
            String formatClassName = displayFormatElement.getAttribute("class");
            Validate.notNull(formatClassName);
            
            try {
                Class formatClass = Class.forName(formatClassName);
                Map parameters = childrenToMap(displayFormatElement);
                
                Format format = getFormatFactory().createFormat(
                        formatClass,
                        getLocale(),
                        parameters);
                
                field.setDisplayFormat(format);
                
            } catch (ClassNotFoundException cnfe) {
            	throw new CayenneRuntimeException(cnfe);
            }
        }

        Element defaultValueElement = getChild(element, "default-value");
        if (defaultValueElement != null) {
            String defaultValueStr = getElementText(defaultValueElement);
            
            Object defaultValue = getDataTypeSpec().create(
                    field.getDataType(),
                    defaultValueStr);
            
            field.setDefaultValue(defaultValue);
        }
    }

    /**
     * Return a Map of child elements of the given element, keyed on element 
     * node name.
     * 
     * @param element the element to get child elements Map from
     * @return the Map of child elements of the given element
     */
    private Map childrenToMap(Element element) {
    	
        List children = getChildren(element);
        
        if (children.isEmpty()) {
            return Collections.EMPTY_MAP;
            
        } else {
            Map map = new HashMap(children.size());
            for (Iterator i = children.iterator(); i.hasNext();) {
                Element child = (Element) i.next();
                map.put(child.getNodeName(), StringUtils.stripToNull(child.getNodeValue()));
            }
            return map;
        }
    }
    
    /**
     * Return the first child element of the given element with the 
     * specified name.
     * 
     * @param element the element to get the child from
     * @param name the NodeName of the element
     * @return the first child elements of the given element and element name
     */
    private Element getChild(Element element, String name) {
    	
        NodeList nodeList = element.getChildNodes();
        
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if (node instanceof Element) {
                if (node.getNodeName().equals(name)) {
                    return (Element) node;
                }
            }
        }
        return null;
    }

    /**
     * Return the list of child elements of the given element with the 
     * specified name.
     * 
     * @param element the element to get the children from
     * @param name the NodeName of the element
     * @return the list of child elements of the given element and element name
     */
    private List getChildren(Element element, String name) {
    	
        List list = new ArrayList();
        NodeList nodeList = element.getChildNodes();
        
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if (node instanceof Element) {
                if (node.getNodeName().equals(name)) {
                    list.add(node);
                }
            }
        }
        return list;
    }

    /**
     * Return the list of child nodes for the given element.
     * 
     * @param element the element to get the child nodes from
     * @return the list of child nodes for the give element
     */
    private List getChildren(Element element) {
    	
        List list = new ArrayList();
        NodeList nodeList = element.getChildNodes();
        
        for (int i = 0; i < nodeList.getLength(); i++) {
            list.add(nodeList.item(i));
        }
        return list;
    }
    
    /**
     * Return the trimmed text value of the given TextNode element. 
     * This method will an empty string if no text value is defined, or if the
     * Node is not an instance of TextNode.
     * 
     * @param element the element to get the text value from
     * @return
     */
    private String getElementText(Element element) {
    	
    	Node node = element.getFirstChild();
    	
    	if (node instanceof TextNode) {
    		return node.toString().trim();
    		
    	} else {
    		return "";
    	}
    }
}