/* ====================================================================
 *
 * The ObjectStyle Group Software License, version 1.1
 * ObjectStyle Group - http://objectstyle.org/
 * 
 * Copyright (c) 2002-2005, Andrei (Andrus) Adamchik and individual authors
 * of the software. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if any,
 *    must include the following acknowlegement:
 *    "This product includes software developed by independent contributors
 *    and hosted on ObjectStyle Group web site (http://objectstyle.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 * 
 * 4. The names "ObjectStyle Group" and "Cayenne" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, email
 *    "andrus at objectstyle dot org".
 * 
 * 5. Products derived from this software may not be called "ObjectStyle"
 *    or "Cayenne", nor may "ObjectStyle" or "Cayenne" appear in their
 *    names without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE OBJECTSTYLE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by many
 * individuals and hosted on ObjectStyle Group web site.  For more
 * information on the ObjectStyle Group, please see
 * <http://objectstyle.org/>.
 */
package org.objectstyle.cayenne.dataview;

import java.util.*;
import java.text.*;

public class FormatFactory {

    private Map builders = new HashMap();

    public FormatFactory() {
        registerBuilder(DecimalFormat.class, new DecimalFormatBuilder());
        registerBuilder(NumberFormat.class, new DecimalFormatBuilder());
        registerBuilder(SimpleDateFormat.class, new SimpleDateFormatBuilder());
        registerBuilder(DateFormat.class, new SimpleDateFormatBuilder());
        registerBuilder(ChoiceFormat.class, new ChoiceFormatBuilder());
        registerBuilder(MessageFormat.class, new MessageFormatBuilder());
        registerBuilder(MapFormat.class, new MapFormatBuilder());
    }

    public Format createFormat(Class formatClass, Locale locale, Map parameters) {
        if (locale == null)
            locale = Locale.getDefault();
        Builder builder = getBuilder(formatClass);
        Format format = builder.create(locale, parameters);
        return format;
    }

    public Builder registerBuilder(Class formatClass, Builder builder) {
        return (Builder) builders.put(formatClass, builder);
    }

    public Builder unregisterBuilder(Class formatClass) {
        return (Builder) builders.remove(formatClass);
    }

    public Builder getBuilder(Class formatClass) {
        return (Builder) builders.get(formatClass);
    }

    public static interface Builder {
        Format create(Locale locale, Map parameters);
    }

    public static class DecimalFormatBuilder implements Builder {
        public Format create(Locale locale, Map parameters) {
            String pattern = (String) parameters.get("pattern");
            DecimalFormatSymbols sym = new DecimalFormatSymbols(locale);
            DecimalFormat format = new DecimalFormat();
            format.setDecimalFormatSymbols(sym);
            if (pattern != null)
                format.applyPattern(pattern);
            return format;
        }
    }

    public static class SimpleDateFormatBuilder implements Builder {
        public Format create(Locale locale, Map parameters) {
            String pattern = (String) parameters.get("pattern");
            DateFormatSymbols sym = new DateFormatSymbols(locale);
            SimpleDateFormat format = new SimpleDateFormat();
            format.setDateFormatSymbols(sym);
            if (pattern != null)
                format.applyPattern(pattern);
            return format;
        }
    }

    public static class ChoiceFormatBuilder implements Builder {
        public Format create(Locale locale, Map parameters) {
            String pattern = (String) parameters.get("pattern");
            ChoiceFormat format = new ChoiceFormat(pattern);
            return format;
        }
    }

    public static class MessageFormatBuilder implements Builder {
        public Format create(Locale locale, Map parameters) {
            String pattern = (String) parameters.get("pattern");
            MessageFormat format = new MessageFormat(pattern, locale);
            return format;
        }
    }

    public static class MapFormatBuilder implements Builder {
        public Format create(Locale locale, Map parameters) {
            MapFormat format = new MapFormat();
            String pattern = (String) parameters.get("pattern");
            String valueClassName = (String) parameters.get("value-class");
            String nullValueDesignation = (String) parameters.get("null-value");
            String entryDelimiter = (String) parameters.get("entry-delimiter");
            String valueFormatDelimiter = (String) parameters
                    .get("value-delimiter");
            if (entryDelimiter != null && entryDelimiter.length() > 0)
                format.setEntryDelimiter(entryDelimiter.charAt(0));
            if (valueFormatDelimiter != null
                    && valueFormatDelimiter.length() > 0)
                format.setValueFormatDelimiter(valueFormatDelimiter.charAt(0));
            if (nullValueDesignation != null)
                format.setNullValueDesignation(nullValueDesignation);
            Class valueClass;
            try {
                valueClass = Class.forName(valueClassName);
                format.applyPattern(pattern, valueClass);
                return format;
            } catch (ClassNotFoundException ex) {
                throw new IllegalArgumentException("Value class "
                        + valueClassName + " not found");
            }
        }
    }
}