/* ====================================================================
 *
 * The ObjectStyle Group Software License, version 1.1
 * ObjectStyle Group - http://objectstyle.org/
 * 
 * Copyright (c) 2002-2005, Andrei (Andrus) Adamchik and individual authors
 * of the software. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if any,
 *    must include the following acknowlegement:
 *    "This product includes software developed by independent contributors
 *    and hosted on ObjectStyle Group web site (http://objectstyle.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 * 
 * 4. The names "ObjectStyle Group" and "Cayenne" must not be used to endorse
 *    or promote products derived from this software without prior written
 *    permission. For written permission, email
 *    "andrus at objectstyle dot org".
 * 
 * 5. Products derived from this software may not be called "ObjectStyle"
 *    or "Cayenne", nor may "ObjectStyle" or "Cayenne" appear in their
 *    names without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE OBJECTSTYLE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by many
 * individuals and hosted on ObjectStyle Group web site.  For more
 * information on the ObjectStyle Group, please see
 * <http://objectstyle.org/>.
 */
package org.objectstyle.cayenne.dataview;

import java.text.Format;

import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.Validate;
import org.objectstyle.cayenne.DataObject;
import org.objectstyle.cayenne.map.ObjAttribute;
import org.objectstyle.cayenne.map.ObjRelationship;

/**
 * Descriptor for a <tt>ObjEntityView</tt> view field.
 * 
 * @see ObjEntityView
 * 
 * @since 1.1
 * @author Andriy Shapochka
 */
public class ObjEntityViewField {

    /** 
     * The field's calculation type, cannot be null. The default value is 
     * <tt>CalcTypeEnum.NO_CALC_TYPE</tt>. 
     */
    private CalcTypeEnum calcType = CalcTypeEnum.NO_CALC_TYPE;

    /**
     * The label caption of this field. Often used in the GUI in a JTable or on 
     * the input form, for example. The default value is an empty string.
     */
    private String caption = "";

    /** The field's data type, cannot be null. */
    private DataTypeEnum dataType = DataTypeEnum.UNKNOWN_TYPE;

    /** The field may have a default value. */
    private Object defaultValue;

    /**
     * The display format of the values for this field. This may differ from 
     * the edit format for example, 1234567.5 dollars can be displayed as 
     * $1,234,567.50, but for edit it is more convenient to accept plain 
     * 1234567.5
     */
    private Format displayFormat;

    /** 
     * The edit format of the values for this field. The default value is null.
     */
    private Format editFormat;

    /** 
     * The actual index of the field in the list of the fields. 
     * The default value is -1.
     */
    private int index = -1;

    /** 
     * If the calc type is lookup then the field must refer to another field
     * to use its values.
     */
    private ObjEntityViewField lookupField;

    /** The field name - unique in the owner's context. */
    private String name = "";

    /** The editability hint to the GUI. */
    private boolean editable = true;

    /**
     * This field maps whether to an ObjAttribute of the ObjEntity
     * the owner (view) corresponds to or an ObjRelationship with
     * the ObjEntity as a source in case this field is a lookup field.
     */
    private ObjAttribute objAttribute;

    private ObjRelationship objRelationship;

    /** ObjEntityView this field belongs to (owner of this field). */
    private ObjEntityView owner;

    /**
     * The preferred index hints how the field should be placed in the ordered
     * list of all the fields of the owner.
     */
    private int preferredIndex = -1;
    
    /** The field is required status. */
    private boolean required;
    
    /** 
     * The is sortable hint, primarily for use in table UI displays. 
     * The default value is false. 
     */
    private boolean sortable;

    /** The field tooltip help message. The default value is null. */
    private String tooltip;
    
    /** The visibility hint to the GUI. */
    private boolean visible = true;
    
    /** The preferred width of the field display. The default value is null. */
    private String width;
    
    // --------------------------------------------------------- Public Methods

    /**
     * Return the <tt>ObjEntityView</tt> owner of the field.
     * 
     * @return the <tt>ObjEntityView</tt> owner of the field
     */
    public ObjEntityView getOwner() {
        return owner;
    }

    /**
     * Return the root <tt>DataView</tt> owner of the field.
     * 
     * @return the root <tt>DataView</tt> owner of the field
     */
    public DataView getRootOwner() {
        return getOwner().getOwner();
    }

    /**
     * The actual index of the field in the list of the fields.
     * 
     * @return the actual index of the field in the list of the fields
     */
    public int getIndex() {
        return index;
    }

    /**
     * Return the preferrred index of the field in the list of fields.
     * 
     * @return the preferred index of the field in the list of fields
     */
    public int getPreferredIndex() {
        return preferredIndex;
    }

    /**
     * Set the preferrred index of the field in the list of fields.
     * 
     * @param preferredIndex the preferred index of the field in the list of fields
     */
    public void setPreferredIndex(int preferredIndex) {
        Validate.isTrue(preferredIndex >= -1);
        this.preferredIndex = preferredIndex;
    }

    /**
     * Return true if the field should be visible.
     * 
     * @return true if the field should be visible
     */
    public boolean isVisible() {
        return visible;
    }

    /**
     * Set whether the field should be visible.
     * 
     * @param visible the field visibility status
     */
    public void setVisible(boolean visible) {
        this.visible = visible;
    }

    /**
     * Return the name of the field.
     * 
     * @return the name of the field
     */
    public String getName() {
        return name;
    }

    /**
     * Set the name of the field.
     * 
     * @param name the name of the field
     */
    public void setName(String name) {
        Validate.notNull(name);
        this.name = name;
    }

    /**
     * Return the field object attribute.
     * 
     * @return the field object attribute
     */
    public ObjAttribute getObjAttribute() {
        return objAttribute;
    }

    /**
     * Set the field object attribute.
     * 
     * @param objAttribute the field object attribute
     */
    public void setObjAttribute(ObjAttribute objAttribute) {
        Validate.notNull(objAttribute);
        this.objAttribute = objAttribute;
    }

    /**
     * Return true if the field is editable.
     * 
     * @return true if the field is editable
     */
    public boolean isEditable() {
        return editable;
    }

    /**
     * Set the field editable status.
     * 
     * @param editable the field editable status
     */
    public void setEditable(boolean editable) {
        this.editable = editable;
    }

    /**
     * Return the label caption of the field.
     * 
     * @return the label caption of the field
     */
    public String getCaption() {
        return caption;
    }

    /**
     * Set the label caption of the field.
     * 
     * @param caption the label caption of the field
     */
    public void setCaption(String caption) {
        this.caption = caption;
    }

    /**
     * Set the <tt>ObjEntityView</tt> owner of the field.
     * 
     * @param owner the <tt>ObjEntityView</tt> owner of the field
     */
    public void setOwner(ObjEntityView owner) {
        this.owner = owner;
    }

    /**
     * Set the index of the field in a list of fields.
     * 
     * @param index the index of the field in the list of fields
     */
    public void setIndex(int index) {
        Validate.isTrue(owner == null || index >= 0);
        this.index = index;
    }

    /**
     * Return the underlying Java Class of the field.
     * 
     * @return the underlying Java Class of the field
     */
    public Class getJavaClass() {
        return getRootOwner().getDataTypeSpec().getJavaClass(dataType);
    }

    /**
     * Return the field value from the given <tt>DataObject</tt>.
     * 
     * @param dataObject the <tt>DataObject</tt> to get the field value from
     * @return the field value from the given data object
     */
    public Object getValue(DataObject dataObject) {
        Object rawValue = getRawValue(dataObject);
        return toValue(rawValue);
    }

    public Object toValue(Object rawValue) {
        Object v = null;
        if (isLookup()) {
            if (rawValue instanceof DataObject)
                v = lookupField.getValue((DataObject) rawValue);
            return v;
        }
        if (rawValue == null)
            return null;

        v = getRootOwner().getDataTypeSpec().toDataType(dataType, rawValue);
        return v;
    }

    public Object toRawValue(Object value) {
        if (value == null)
            return null;
        DataView rootOwner = getOwner().getOwner();
        if (isLookup()) {
            return rootOwner.getLookupCache().getDataObject(lookupField, value);
        }
        if (objAttribute == null)
            return null;

        String type = objAttribute.getType();
        Object v = null;
        try {
            Class untypedValueClass = Class.forName(type);
            v = rootOwner.getDataTypeSpec().fromDataType(
                    untypedValueClass,
                    dataType,
                    value);
        }
        catch (ClassNotFoundException ex) {
            ex.printStackTrace();
        }
        return v;
    }

    public void setValue(DataObject obj, Object value) {
        Object rawValue = toRawValue(value);
        Object oldValue = getValue(obj);
        setRawValue(obj, rawValue);
        getRootOwner().fireFieldValueChangeEvent(this, obj, oldValue, value);
    }

    public String getFormattedValue(DataObject obj) {
        Object value = getRawValue(obj);
        String formattedValue = null;
        if (!isLookup()) {
            Format f = (displayFormat != null ? displayFormat : editFormat);
            if (f == null)
                formattedValue = ObjectUtils.toString(value);
            else {
                try {
                    formattedValue = f.format(value);
                }
                catch (Exception ex) {
                    formattedValue = "";
                }
            }
        }
        else {
            formattedValue = lookupField.getFormattedValue((DataObject) value);
        }
        return formattedValue;
    }

    public Object getRawValue(DataObject obj) {
        if (obj == null)
            return null;
        Object value = null;
        if (!isLookup() && objAttribute != null) {
            value = obj.readProperty(objAttribute.getName());
        }
        else if (isLookup() && objRelationship != null) {
            value = obj.readProperty(objRelationship.getName());
        }
        return value;
    }

    public void setRawValue(DataObject obj, Object value) {
        if (obj != null) {
            if (!isLookup() && objAttribute != null) {
                obj.writeProperty(objAttribute.getName(), value);
            }
            else if (isLookup() && objRelationship != null) {
                obj.setToOneTarget(
                        objRelationship.getName(),
                        (DataObject) value,
                        objRelationship.getReverseRelationship() != null);
            }
        }
    }
    
    /**
     * Return the field <tt>DataTypeEnum</tt> data type.
     * 
     * @return the field <tt>DataTypeEnum</tt> data type
     */
    public DataTypeEnum getDataType() {
        return dataType;
    }
    
    /**
     * Return the field <tt>DataTypeEnum</tt> data type.
     * 
     * @return the field <tt>DataTypeEnum</tt> data type
     */
    public void setDataType(DataTypeEnum dataType) {
        Validate.notNull(dataType);
        this.dataType = dataType;
    }

    public CalcTypeEnum getCalcType() {
        return calcType;
    }

    public void setCalcType(CalcTypeEnum calcType) {
        Validate.notNull(calcType);
        this.calcType = calcType;
    }

    public Object getDefaultValue() {
        return defaultValue;
    }

    public void setDefaultValue(Object defaultValue) {
        this.defaultValue = defaultValue;
    }

    public ObjRelationship getObjRelationship() {
        return objRelationship;
    }

    public void setObjRelationship(ObjRelationship objRelationship) {
        Validate.notNull(objRelationship);
        this.objRelationship = objRelationship;
    }

    public boolean isLookup() {
        return CalcTypeEnum.LOOKUP_TYPE.equals(calcType);
    }

    public Object[] getLookupValues() {
        if (!isLookup())
            return null;
        return getRootOwner().getLookupCache().getCachedValues(lookupField);
    }

    public ObjEntityViewField getLookupField() {
        return lookupField;
    }

    public void setLookupField(ObjEntityViewField lookupField) {
        this.lookupField = lookupField;
    }

    public Format getDisplayFormat() {
        if (displayFormat == null
                && isLookup()
                && lookupField != null
                && lookupField != this)
            return lookupField.getDisplayFormat();
        return displayFormat;
    }

    public void setDisplayFormat(Format displayFormat) {
        this.displayFormat = displayFormat;
    }

    public Format getEditFormat() {
        if (editFormat == null
                && isLookup()
                && lookupField != null
                && lookupField != this)
            return lookupField.getEditFormat();
        return editFormat;
    }

    public void setEditFormat(Format editFormat) {
        this.editFormat = editFormat;
    }

    public boolean isSameObjAttribute(ObjEntityViewField field) {
        if (objAttribute != null) {
            ObjAttribute fieldAttribute = field.getObjAttribute();
            if (fieldAttribute == null)
                return false;
            boolean same = (objAttribute.equals(fieldAttribute) || (objAttribute
                    .getEntity()
                    .equals(fieldAttribute.getEntity()) && objAttribute.getName().equals(
                    fieldAttribute.getName())));
            return same;
        }
        else if (isLookup()) {
            if (field.isLookup())
                return getLookupField().isSameObjAttribute(field.getLookupField());
            else
                return getLookupField().isSameObjAttribute(field);
        }
        return false;
    }

    /**
     * Return the field required status.
     * 
     * @return the field required status
     */
	public boolean isRequired() {
		return required;
	}

	/**
	 * Set the field required status.
	 * 
	 * @param required the field required status
	 */
	public void setRequired(boolean required) {
		this.required = required;
	}

	/**
	 * Return the field sortable UI hint.
	 * 
	 * @return the field sortable UI hint.
	 */
	public boolean isSortable() {
		return sortable;
	}

	/**
	 * Set the field sortable UI hint.
	 * 
	 * @param sortable the field sortable UI hint
	 */
	public void setSortable(boolean sortable) {
		this.sortable = sortable;
	}

	/**
	 * Return the preferrred width of the field display.
	 * 
	 * @return the preferrred width of the field display
	 */
	public String getWidth() {
		return width;
	}

	/**
	 * Set the preferrred width of the field display.
	 * 
	 * @param width the preferrred width of the field display
	 */
	public void setWidth(String width) {
		this.width = width;
	}

	/**
	 * Return the field tooltip help message. 
	 * 
	 * @return the field tooltip help message
	 */
	public String getTooltip() {
		return tooltip;
	}

	/**
	 * Set the field tooltip help message.
	 * 
	 * @param tooltip the field tooltip help message
	 */
	public void setTooltip(String tooltip) {
		this.tooltip = tooltip;
	}

}